@@ -39,12 +39,19 @@ struct rktio_fd_t {
   struct Win_FD_Input_Thread *th; /* input mode */
   struct Win_FD_Output_Thread *oth; /* output mode */
   int unblocked; /* whether non-blocking mode is installed */
+  int write_limit; /* non-0 => max on amount to try writing */
   char *buffer; /* shared with reading thread */
   int has_pending_byte; /* for text-mode input, may be dropped by a following lf */
   int pending_byte; /* for text-mode input, either a CR waiting to decode, or byte that didn't fit */
+  int leftover_len; /* for bytes that should be written, but that form a UTF-8 encoding prefix */
+  char leftover[6];
 #endif
 };
 
+struct rktio_fd_transfer_t {
+  rktio_fd_t fd;
+};
+
 /*========================================================================*/
 /* Windows I/O helper structs                                             */
 /*========================================================================*/
@@ -87,15 +94,18 @@ typedef struct Win_FD_Output_Thread {
 # define RKTIO_FD_BUFFSIZE 4096
 
 static void init_read_fd(rktio_t *rktio, rktio_fd_t *rfd);
-static void deinit_read_fd(rktio_t *rktio, rktio_fd_t *rfd, int full_close);
-static void deinit_write_fd(rktio_t *rktio, rktio_fd_t *rfd, int full_close);
+static void deinit_read_fd(rktio_fd_t *rfd, int full_close);
+static void deinit_write_fd(rktio_fd_t *rfd, int full_close);
 
-static void deinit_fd(rktio_t *rktio, rktio_fd_t *rfd, int full_close)
+static void deinit_fd(rktio_fd_t *rfd, int full_close)
 {
-  deinit_read_fd(rktio, rfd, full_close);
-  deinit_write_fd(rktio, rfd, full_close);
+  deinit_read_fd(rfd, full_close);
+  deinit_write_fd(rfd, full_close);
 }
 
+#define DELAYED_CONSOLE_HANDLE ((HANDLE)-2)
+static void force_console(rktio_fd_t *rfd);
+
 static long WINAPI WindowsFDReader(Win_FD_Input_Thread *th);
 static void WindowsFDICleanup(Win_FD_Input_Thread *th, int close_mode);
 
@@ -128,6 +138,11 @@ static intptr_t adjust_input_text_for_pending_cr(rktio_fd_t *rfd, char *buffer,
 static const char *adjust_output_text(const char *buffer, intptr_t *towrite);
 static intptr_t recount_output_text(const char *orig_buffer, const char *buffer, intptr_t wrote);
 
+static wchar_t *convert_output_wtext(const char *buffer, intptr_t *_towrite,
+				     int *_can_leftover, int *_keep_leftover,
+				     int leftover_len, char *leftover);
+static intptr_t recount_output_wtext(wchar_t *w_buffer, intptr_t winwrote);
+
 #endif
 
 /*========================================================================*/
@@ -162,7 +177,8 @@ rktio_fd_t *rktio_system_fd(rktio_t *rktio, intptr_t system_fd, int modes)
     rfd->sock = (SOCKET)system_fd;
   else
     rfd->fd = (HANDLE)system_fd;
-  if (!(modes & (RKTIO_OPEN_REGFILE | RKTIO_OPEN_NOT_REGFILE | RKTIO_OPEN_SOCKET))) {
+  if (!(modes & (RKTIO_OPEN_REGFILE | RKTIO_OPEN_NOT_REGFILE | RKTIO_OPEN_SOCKET))
+      && (rfd->fd != DELAYED_CONSOLE_HANDLE)) {
     if ((GetFileType(rfd->fd) == FILE_TYPE_DISK))
       rfd->modes |= RKTIO_OPEN_REGFILE;
     if (!(modes & (RKTIO_OPEN_DIR | RKTIO_OPEN_NOT_DIR))) {
@@ -185,7 +201,7 @@ rktio_fd_t *rktio_system_fd(rktio_t *rktio, intptr_t system_fd, int modes)
   return rfd;
 }
 
-intptr_t rktio_fd_system_fd(rktio_t *rktio, rktio_fd_t *rfd)
+intptr_t rktio_internal_fd_system_fd(rktio_fd_t *rfd)
 {
 #ifdef RKTIO_SYSTEM_UNIX
   return rfd->fd;
@@ -193,11 +209,18 @@ intptr_t rktio_fd_system_fd(rktio_t *rktio, rktio_fd_t *rfd)
 #ifdef RKTIO_SYSTEM_WINDOWS
   if (rfd->modes & RKTIO_OPEN_SOCKET)
     return (intptr_t)rfd->sock;
-  else
+  else {
+    force_console(rfd);
     return (intptr_t)rfd->fd;
+  }
 #endif
 }
 
+intptr_t rktio_fd_system_fd(rktio_t *rktio, rktio_fd_t *rfd)
+{
+  return rktio_internal_fd_system_fd(rfd);
+}
+
 rktio_fd_t *rktio_std_fd(rktio_t *rktio, int which)
 {
   int mode = ((which == RKTIO_STDIN)
@@ -207,6 +230,7 @@ rktio_fd_t *rktio_std_fd(rktio_t *rktio, int which)
   return rktio_system_fd(rktio, which, mode | RKTIO_OPEN_NOT_DIR);
 #endif
 #ifdef RKTIO_SYSTEM_WINDOWS
+  HANDLE h;
   switch (which) {
   case RKTIO_STDIN:
     which = STD_INPUT_HANDLE;
@@ -218,8 +242,12 @@ rktio_fd_t *rktio_std_fd(rktio_t *rktio, int which)
     which = STD_ERROR_HANDLE;
     break;
   }
+  h = GetStdHandle(which);
+  if ((h == INVALID_HANDLE_VALUE) || (h == NULL)) {
+    h = DELAYED_CONSOLE_HANDLE; /* => open a console on demand */
+  }
   return rktio_system_fd(rktio,
-			 (intptr_t)GetStdHandle(which),
+			 (intptr_t)h,
 			 mode | RKTIO_OPEN_NOT_DIR);
 #endif
 }
@@ -255,7 +283,9 @@ int rktio_system_fd_is_terminal(rktio_t *rktio, intptr_t fd)
   return isatty(fd);
 #endif
 #ifdef RKTIO_SYSTEM_WINDOWS
-  if (GetFileType((HANDLE)fd) == FILE_TYPE_CHAR) {
+  if ((HANDLE)fd == DELAYED_CONSOLE_HANDLE)
+    return 1; /* delayed console */
+  else if (GetFileType((HANDLE)fd) == FILE_TYPE_CHAR) {
     DWORD mode;
     if (GetConsoleMode((HANDLE)fd, &mode))
       return 1;
@@ -307,9 +337,13 @@ rktio_fd_t *rktio_dup(rktio_t *rktio, rktio_fd_t *rfd)
     HANDLE  newhandle;
     BOOL rc;
 
-    rc = DuplicateHandle(GetCurrentProcess(), rfd->fd,
-                         GetCurrentProcess(), &newhandle,
-                         0, FALSE, DUPLICATE_SAME_ACCESS);
+    if (rfd->fd == DELAYED_CONSOLE_HANDLE) {
+      newhandle = DELAYED_CONSOLE_HANDLE;
+      rc = TRUE;
+    } else
+      rc = DuplicateHandle(GetCurrentProcess(), rfd->fd,
+			   GetCurrentProcess(), &newhandle,
+			   0, FALSE, DUPLICATE_SAME_ACCESS);
 
     if (rc == FALSE) {
       get_windows_error();
@@ -340,20 +374,20 @@ void rktio_reliably_close(intptr_t s)
 }
 #endif
 
-static rktio_ok_t do_close(rktio_t *rktio, rktio_fd_t *rfd, int set_error)
+static rktio_ok_t do_close(rktio_t *rktio /* maybe NULL */, rktio_fd_t *rfd, int set_error)
 {
   int ok;
 
 #ifdef RKTIO_SYSTEM_UNIX
   int cr;
 
+  cr = rktio_reliably_close_err(rfd->fd);
+
 # ifdef RKTIO_USE_FCNTL_AND_FORK_FOR_FILE_LOCKS
-  if (!(rfd->modes & RKTIO_OPEN_SOCKET))
+  if (rktio && !(rfd->modes & RKTIO_OPEN_SOCKET))
     rktio_release_lockf(rktio, rfd->fd);
 # endif
 
-  cr = rktio_reliably_close_err(rfd->fd);
-
   if (cr && set_error) {
     get_posix_error();   
     ok = 0;
@@ -364,21 +398,22 @@ static rktio_ok_t do_close(rktio_t *rktio, rktio_fd_t *rfd, int set_error)
   if (rfd->modes & RKTIO_OPEN_SOCKET)
     return rktio_socket_close(rktio, rfd, set_error);
 
-  deinit_fd(rktio, rfd, 1);
+  deinit_fd(rfd, 1);
 
   ok = 1;
-  if (!rfd->th && !rfd->oth) {
+  if (!rfd->th && !rfd->oth && (rfd->fd != DELAYED_CONSOLE_HANDLE)) {
     if (!CloseHandle(rfd->fd)) {
       ok = 0;
-      get_windows_error();
+      if (set_error)
+        get_windows_error();
     }
   }
   
 #endif
 
-  if (ok)
+  if (ok || !set_error)
     free(rfd);
-  
+
   return ok;
 }
 
@@ -392,14 +427,38 @@ void rktio_close_noerr(rktio_t *rktio, rktio_fd_t *rfd)
   (void)do_close(rktio, rfd, 0);
 }
 
+rktio_ok_t rktio_internal_close(rktio_t *rktio /* may be NULL */, rktio_fd_t *rfd, int set_error)
+{
+  return do_close(rktio, rfd, set_error);
+}
+
 void rktio_forget(rktio_t *rktio, rktio_fd_t *rfd)
 {
 #ifdef RKTIO_WINDOWS_SYSTEM
-  deinit_fd(rktio, rfd, 1);
+  deinit_fd(rfd, 1);
 #endif
   free(rfd);
 }
 
+/*========================================================================*/
+/* detach, attach, and abandon                                            */
+/*========================================================================*/
+
+rktio_fd_transfer_t *rktio_fd_detach(rktio_t *rktio, rktio_fd_t *rfd)
+{
+  return (rktio_fd_transfer_t *)rfd;
+}
+
+rktio_fd_t *rktio_fd_attach(rktio_t *rktio, rktio_fd_transfer_t *rfdt)
+{
+  return (rktio_fd_t *)rfdt;
+}
+
+void rktio_fd_close_transfer(rktio_fd_transfer_t *rfdt)
+{
+  (void)do_close(NULL, (rktio_fd_t *)rfdt, 0);
+}
+
 /*========================================================================*/
 /* polling                                                                */
 /*========================================================================*/
@@ -570,6 +629,9 @@ int poll_write_ready_or_flushed(rktio_t *rktio, rktio_fd_t *rfd, int check_flush
   }
 #endif
 #ifdef RKTIO_SYSTEM_WINDOWS
+  if (rfd->fd == DELAYED_CONSOLE_HANDLE)
+    return RKTIO_POLL_READY;
+
   if (rfd->modes & RKTIO_OPEN_SOCKET) {
     if (check_flushed)
       return RKTIO_POLL_READY;
@@ -656,6 +718,8 @@ void rktio_poll_add(rktio_t *rktio, rktio_fd_t *rfd, rktio_poll_set_t *fds, int
     fds2 = RKTIO_GET_FDSET(fds, 2);
     RKTIO_FD_SET((intptr_t)rfd->sock, fds2);
   } else {
+    force_console(rfd);
+
     if (modes & RKTIO_POLL_READ) {
       init_read_fd(rktio, rfd);
       if (rfd->th) {
@@ -740,15 +804,21 @@ intptr_t rktio_read_converted(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, int
     else
       return bc;
   } else {
-    /* We use a non-blocking read here, even though we've waited
-       for input above, because an external process might have
-       gobbled the characters that we expected to get. */
     int old_flags;
+
+    /* Check that input is available. This step is needed
+       for fifos to avoid getting an EOF-looking result
+       if there's no writer, yet. */
+    if (!rktio_poll_read_ready(rktio, rfd))
+      return 0;
     
     old_flags = fcntl(rfd->fd, F_GETFL, 0);
     if (!(old_flags & RKTIO_NONBLOCKING))
       fcntl(rfd->fd, F_SETFL, old_flags | RKTIO_NONBLOCKING);
     
+    /* We use a non-blocking read here, even though we've waited
+       for input above, because an external process might have
+       gobbled the characters that we expected to get. */
     do {
       bc = read(rfd->fd, buffer, len);
     } while ((bc == -1) && errno == EINTR);
@@ -797,6 +867,7 @@ intptr_t rktio_read_converted(rktio_t *rktio, rktio_fd_t *rfd, char *buffer, int
 	}
       } else {
 	/* read after first byte installed into the buffer */
+        rfd->has_pending_byte = 0;
 	offset = 1;
       }
     }
@@ -937,6 +1008,8 @@ static intptr_t adjust_input_text_for_pending_cr(rktio_fd_t *rfd, char *buffer,
 
 static void init_read_fd(rktio_t *rktio, rktio_fd_t *rfd)
 {
+  force_console(rfd);
+
   if (!rktio_fd_is_regular_file(rktio, rfd) && !rfd->th) {
     /* To get non-blocking I/O for anything that can block, we create
        a separate reader thread.
@@ -979,7 +1052,7 @@ static void init_read_fd(rktio_t *rktio, rktio_fd_t *rfd)
   }
 }
 
-static void deinit_read_fd(rktio_t *rktio, rktio_fd_t *rfd, int full_close)
+static void deinit_read_fd(rktio_fd_t *rfd, int full_close)
 {
   if (rfd->th) {
     CSI_proc csi;
@@ -1019,15 +1092,21 @@ static void deinit_read_fd(rktio_t *rktio, rktio_fd_t *rfd, int full_close)
 static long WINAPI WindowsFDReader(Win_FD_Input_Thread *th)
 {
   DWORD toget, got;
-  int perma_eof = 0;
+  int perma_eof = 0, ft, is_console = 0;
   HANDLE eof_wait = NULL;
+# define CONSOLE_BUFFER_IN_SIZE 16
+  wchar_t w_buffer[CONSOLE_BUFFER_IN_SIZE];
+  int w_buf_count = 0;
 
-  if (GetFileType((HANDLE)th->fd) == FILE_TYPE_PIPE) {
+  ft = GetFileType((HANDLE)th->fd);
+    
+  if (ft == FILE_TYPE_PIPE) {
     /* Reading from a pipe will return early when data is available. */
     toget = RKTIO_FD_BUFFSIZE;
   } else {
     /* Non-pipe: get one char at a time: */
     toget = 1;
+    is_console = (ft == FILE_TYPE_CHAR);
   }
 
   while (!perma_eof && !th->err) {
@@ -1042,8 +1121,39 @@ static long WINAPI WindowsFDReader(Win_FD_Input_Thread *th)
       /* Spurious wake-up? */
       ReleaseSemaphore(th->lock_sema, 1, NULL);
     } else {
+      int ok;
       ReleaseSemaphore(th->lock_sema, 1, NULL);
-      if (ReadFile(th->fd, th->buffer, toget, &got, NULL)) {
+      if (!is_console)
+	ok = ReadFile(th->fd, th->buffer, toget, &got, NULL);
+      else {
+	if (w_buf_count) {
+	  ok = 1;
+	  got = w_buf_count;
+	} else
+	  ok = ReadConsoleW(th->fd, w_buffer, CONSOLE_BUFFER_IN_SIZE, &got, NULL);
+	if (ok) {
+	  /* check for Ctl-Z, and convert it to an EOF */
+	  int i, move_start = 0, move_len = 0;
+	  for (i = 0; i < got; i++) {
+	    if (w_buffer[i] == 26 /* ctl-z */) {
+	      move_start = i;
+	      move_len = got - i;
+	      if (i == 0) {
+		/* report EOF now */
+		move_start++;
+		move_len--;
+	      }
+	      got = i;
+	      break;
+	    }
+	  }
+	  got = WideCharToMultiByte(CP_UTF8, 0, w_buffer, got, th->buffer, RKTIO_FD_BUFFSIZE, NULL, 0);
+	  if (move_len > 0)
+	    memmove(w_buffer, w_buffer + move_start, sizeof(wchar_t) * move_len);
+	  w_buf_count = move_len;
+	}
+      }
+      if (ok) {
 	WaitForSingleObject(th->lock_sema, INFINITE);
 	th->avail = got;
 	th->offset = 0;
@@ -1163,26 +1273,51 @@ intptr_t rktio_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr
   if (rfd->modes & RKTIO_OPEN_SOCKET)
     return rktio_socket_write(rktio, rfd, buffer, len);
 
+  force_console(rfd);
+
   if (rktio_fd_is_regular_file(rktio, rfd)
       || rktio_fd_is_terminal(rktio, rfd)) {
     /* Regular files never block, so this code looks like the Unix
-       code.   */
+       code. */
 
     /* If we try to write too much at once, the result
        is ERROR_NOT_ENOUGH_MEMORY (as opposed to a partial write). */
-    int ok;
+    int ok, to_console, can_leftover = 0, keep_leftover = 0;
     intptr_t towrite = len;
     const char *orig_buffer = buffer;
+    wchar_t *w_buffer = NULL;
+    DWORD max_winwrote;
     int err;
 
     if (rfd->modes & RKTIO_OPEN_TEXT)
       buffer = adjust_output_text(buffer, &towrite);
+
+    max_winwrote = towrite;
+  
+    to_console = rktio_fd_is_terminal(rktio, rfd);
+    if (to_console) {
+      /* Decode UTF-8 and write a chunk on a character boundary. */
+      w_buffer = convert_output_wtext(buffer, &towrite,
+				      &can_leftover, &keep_leftover,
+				      rfd->leftover_len, rfd->leftover);
+    }
     
     while (1) {
-      ok = WriteFile((HANDLE)rfd->fd, buffer, towrite, &winwrote, NULL);
+      if (!to_console)
+	ok = WriteFile((HANDLE)rfd->fd, buffer, towrite, &winwrote, NULL);
+      else {
+	if (towrite)
+	  ok = WriteConsoleW((HANDLE)rfd->fd, w_buffer, towrite, &winwrote, NULL);
+	else {
+	  /* can happend if can_leftover is > 0 */
+	  ok = 1;
+	  winwrote = 0;
+	}
+      }
+
       if (!ok)
         err = GetLastError();
-	    
+
       if (!ok && (err == ERROR_NOT_ENOUGH_MEMORY)) {
         towrite = towrite >> 1;
         if (towrite && (buffer != orig_buffer)) {
@@ -1203,6 +1338,30 @@ intptr_t rktio_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr
       return RKTIO_WRITE_ERROR;
     }
 
+    if (to_console) {
+      /* Convert wchar count to byte count, taking into account leftovers */
+      int wrote_all = (winwrote == towrite);
+      if (winwrote) {
+	/* Recounting only works right if the outptu was well-formed
+	   UTF-8. Weird things happen otherwise... but we guard against
+	   external inconsistency with the `max_winwrote` check below. */
+	winwrote = recount_output_wtext(w_buffer, winwrote);
+	if (winwrote > rfd->leftover_len)
+	  winwrote -= rfd->leftover_len;
+	else
+	  winwrote = 0;
+	rfd->leftover_len = 0;
+      }
+      if (wrote_all && can_leftover) {
+	memcpy(rfd->leftover + keep_leftover, buffer + max_winwrote - can_leftover, can_leftover);
+	rfd->leftover_len = can_leftover + keep_leftover;
+	winwrote += can_leftover;
+      }
+      free(w_buffer);
+      if (winwrote > max_winwrote)
+	winwrote = max_winwrote;
+    }
+
     if (buffer != orig_buffer) {
       /* Convert converted count back to original count: */
       winwrote = recount_output_text(orig_buffer, buffer, winwrote);
@@ -1234,7 +1393,6 @@ intptr_t rktio_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr
       if (nonblocking) {
         /* Unless we're still trying to flush old data, write to the
            pipe and have the other thread start flushing it. */
-        DWORD nonblock = PIPE_NOWAIT;
         int flushed;
 
         if (rfd->oth) {
@@ -1266,13 +1424,34 @@ intptr_t rktio_write(rktio_t *rktio, rktio_fd_t *rfd, const char *buffer, intptr
              write fails. (Yuck.) */
           while (1) {
             if (!rfd->unblocked) {
+	      DWORD nonblock = PIPE_NOWAIT;
               ok = SetNamedPipeHandleState((HANDLE)rfd->fd, &nonblock, NULL, NULL);
               if (ok)
                 rfd->unblocked = 1;
-              else
+              else {
                 errsaved = GetLastError();
+		if (errsaved == ERROR_INVALID_FUNCTION) {
+		  /* The handle (not a pipe?) doesn't support non-blocking mode. But
+		     since we only try to write when the pipe is flushed, we can just
+		     keep each request under the buffer size. */
+		  DWORD bufsz;
+		  if (GetNamedPipeInfo((HANDLE)rfd->fd, NULL, &bufsz, NULL, NULL)
+		      && (bufsz > 0))
+		    rfd->write_limit = bufsz;
+		  else {
+		    /* 256 should be small enough? */
+		    rfd->write_limit = 256;
+		  }
+		  rfd->unblocked = 1;
+		  ok = 1;
+		} 
+	      }
             } else
               ok = 1;
+
+	    if (rfd->write_limit && (towrite > rfd->write_limit))
+	      towrite = rfd->write_limit;
+
             if (ok) {
               ok = WriteFile((HANDLE)rfd->fd, buffer, towrite, &winwrote, NULL);
               if (!ok)
@@ -1462,7 +1641,118 @@ static intptr_t recount_output_text(const char *orig_buffer, const char *buffer,
   return i;
 }
 
-static void deinit_write_fd(rktio_t *rktio, rktio_fd_t *rfd, int full_close)
+static wchar_t *convert_output_wtext(const char *buffer, intptr_t *_towrite,
+				     int *_can_leftover, int *_keep_leftover,
+				     int leftover_len, char *leftover)
+{
+  /* Figure out how many bytes we can convert to complete wide
+     characters. To avoid quadratic behavior overall, we'll limit the
+     number of bytes.
+
+     The given `leftover_len` and `leftover` is a prefix on `buffer`.
+     If the tail (after writing all other bytes) is an incomplete
+     UTF-8 prefix, report the prefix length in `_can_leftover`. */
+  intptr_t i, count, len = *_towrite;
+  char *src_buffer;
+  wchar_t *dest_buffer;
+  int want, span = 0;
+
+  if (leftover_len) {
+    /* Assume that leftover is a valid prefix: */
+    int v = ((unsigned char *)leftover)[0];
+    if ((v & 0xF8) == 0xF0)
+      span = 4;
+    else if ((v & 0xF0) == 0xE0)
+      span = 3;
+    else
+      span = 2;
+  } else
+    span = 0;
+  want = span - leftover_len;
+
+  for (i = 0, count = 0; (i < len) && (count < 1024); i++) {
+    int v = ((unsigned char *)buffer)[i];
+    if (want) {
+      if ((v & 0xC0) == 0x80) {
+	/* valid continuation byte */
+	want--;
+	if (!want) {
+	  count++;
+	  if (span == 4)
+	    count++; /* surrogate pair */
+	}
+      } else {
+	/* not a valid continuation byte */
+	count++;
+	want = 0;
+	--i; /* retry byte */
+      }
+    } else if (!(v & 0x80)) {
+      count++;
+    } else if ((v & 0xF8) == 0xF0) {
+      span = 4;
+      want = 3;
+    } else if ((v & 0xF0) == 0xE0) {
+      span = 3;
+      want = 2;
+    } else if ((v & 0xE0) == 0xC0) {
+      span = 2;
+      want = 1;
+    } else {
+      /* bad continuation byte */
+      count++;
+    }
+  }
+
+  if ((i == len) && (want > 0)) {
+    /* consuming all input, so set leftover */
+    int keep = span - want;
+    if (i >= keep) {
+      *_can_leftover = keep;
+      i -= keep;
+      *_keep_leftover = 0;
+    } else {
+      *_can_leftover = (keep - leftover_len);
+      *_keep_leftover = leftover_len;
+      i = 0;
+    }
+  } else {
+    *_can_leftover = 0;
+    *_keep_leftover = 0;
+  }
+
+  if (leftover_len) {
+    src_buffer = malloc(i + leftover_len);
+    memcpy(src_buffer, leftover, leftover_len);
+    memcpy(src_buffer + leftover_len, buffer, i);
+    i += leftover_len;
+  } else
+    src_buffer = (char *)buffer;
+
+  dest_buffer = (wchar_t *)malloc(sizeof(wchar_t) * count);
+
+  if (count > 0) {
+    count = MultiByteToWideChar(CP_UTF8, 0, src_buffer, i, dest_buffer, count);
+    if (!count) {
+      /* force progress */
+      src_buffer[0] = 0xFFFD;
+      count = 1;
+    }
+  }
+  *_towrite = count;
+
+  if (leftover_len)
+    free(src_buffer);
+
+  return dest_buffer;
+}
+
+static intptr_t recount_output_wtext(wchar_t *w_buffer, intptr_t winwrote)
+{
+  return WideCharToMultiByte(CP_UTF8, 0, w_buffer, winwrote, NULL, 0, NULL, 0);
+}
+
+static void deinit_write_fd(rktio_fd_t *rfd, int full_close)
 {
   if (rfd->oth) {
     CSI_proc csi;
@@ -1534,9 +1824,9 @@ static long WINAPI WindowsFDWriter(Win_FD_Output_Thread *oth)
 	err_no = 0;
 
       WaitForSingleObject(oth->lock_sema, INFINITE);
-      if (!ok)
+      if (!ok) {
 	oth->err_no = err_no;
-      else {
+      } else {
 	oth->buflen -= wrote;
 	if (oth->buflen)
 	  memmove(oth->buffer, oth->buffer + wrote, oth->buflen);
@@ -1574,3 +1864,29 @@ static void WindowsFDOCleanup(Win_FD_Output_Thread *oth, int close_mode)
 }
 
 #endif
+
+/*========================================================================*/
+/* console                                                                */
+/*========================================================================*/
+
+#ifdef RKTIO_SYSTEM_WINDOWS
+
+static void force_console(rktio_fd_t *rfd) {
+  /* DELAYED_CONSOLE_HANDLE is used to indicate that a console should be created on demand */
+  if (rfd->fd == DELAYED_CONSOLE_HANDLE) {
+    HANDLE h;
+    int which;
+
+    rktio_create_console();
+
+    if (rfd->modes & RKTIO_OPEN_READ)
+      which = STD_INPUT_HANDLE;
+    else
+      which = STD_OUTPUT_HANDLE;
+    h = GetStdHandle(which);
+
+    rfd->fd = h;
+  }
+}
+
+#endif
